from typing import List
from dataclasses import dataclass


class Migration:
    warnings: List[str]

    @staticmethod
    def target_version() -> str:
        raise NotImplementedError

    @staticmethod
    def source_version() -> str:
        raise NotImplementedError

    def __init__(self, data: dict) -> None:
        self.data = data
        self.warnings = []

    @property
    def data(self) -> dict:
        raise NotImplementedError

    @data.setter
    def data(self, data: dict):
        self.data = data

    def _migrate(self) -> None:
        raise NotImplementedError

    def _bump_version(self) -> None:
        self.data["version"] = self.target_version

    def apply(self) -> None:
        self._migrate()
        self._bump_version()



import flask, flask_sock

from ...settings import Settings
from ..utils import send_from_dist
from ...execution.execution import RequestData
from ...execution import HookExecution, JobExecution, FormExecution
from ...utils.environment import BUILD_ID, SIDECAR_SHARED_TOKEN, SHOW_WATERMARK

from ..controller.main import MainController
from ..controller import auth as auth_controller


def get_player_bp(controller: MainController):
    bp = flask.Blueprint("player", __name__)
    sock = flask_sock.Sock(bp)

    auth_bp = auth_controller.get_player_bp()
    bp.register_blueprint(auth_bp, url_prefix="/_auth")

    @bp.route("/_workspace", methods=["GET"])
    def _get_workspace():
        workspace = controller.get_workspace()
        return workspace.browser_runner_dto

    @bp.route("/_pages/<string:path>", methods=["GET"])
    def _get_page(path):
        page_runtime = controller.get_form_by_path(path)
        if not page_runtime:
            flask.abort(404)

        return {
            page_runtime.runner_type: {
                **page_runtime.browser_runner_dto,
                "workspace": controller.get_workspace().browser_runner_dto,
            }
        }

    @bp.route("/_version", methods=["GET"])
    def _get_version():
        return BUILD_ID

    @bp.route("/_settings", methods=["GET"])
    def _get_settings():
        return flask.jsonify({"show_watermark": SHOW_WATERMARK})

    @sock.route("/_socket")
    def _websocket(conn: flask_sock.Server):
        request_data = RequestData(
            query_params=flask.request.args,
            body=flask.request.get_data(as_text=True),
            headers=flask.request.headers,
            method=flask.request.method,
        )

        try:
            id = flask.request.args.get("id")
            if id is None:
                return conn.close(reason=400, message="No path")

            form = controller.get_form(id)
            if not form:
                return conn.close(reason=404, message="Not found")

            execution = FormExecution(
                is_initial=controller.is_initial(form.id),
                request=request_data,
                runtime_json=form,
                connection=conn,
            )

            execution.run_sync()
            controller.run_waiting_scripts(execution.stage_run)
        finally:
            conn.close(message="Done")

    @bp.route("/_files", methods=["PUT"])
    def _upload_file():
        files = flask.request.files
        filename = flask.request.form.get("filename")
        if filename is None:
            flask.abort(400)

        if len(files) == 0:
            flask.abort(400)

        paths = []
        for file in files.values():
            paths.append(controller.save_file(file, filename))
        return paths

    @bp.route("/_files/<path:path>", methods=["GET"])
    def _get_file(path):
        return flask.send_file(controller.get_file(path))

    @bp.route("/_assets/logo", methods=["GET"])
    def _logo():
        logo_path = controller.get_workspace().logo_url
        if not logo_path:
            return flask.abort(404)
        return send_from_dist(logo_path, dist_folder=Settings.root_path)

    @bp.route("/favicon.ico", methods=["GET"])
    def _favicon():
        return _logo()

    @bp.route("/_assets/background", methods=["GET"])
    def _background():
        background_path = controller.get_workspace().theme

        if not background_path:
            return flask.abort(404)

        return send_from_dist(background_path, dist_folder=Settings.root_path)

    @bp.route("/_hooks/<path:path>", methods=["POST", "GET", "PUT", "DELETE", "PATCH"])
    def hook_runner(path):
        hook = controller.get_hook_by_path(path)

        if not hook:
            flask.abort(404)

        if not hook.file:
            flask.abort(500)

        request_data = RequestData(
            query_params=flask.request.args,
            body=flask.request.get_data(as_text=True),
            headers=flask.request.headers,
            method=flask.request.method,
        )

        execution = HookExecution(
            is_initial=controller.is_initial(hook.id),
            request=request_data,
            runtime_json=hook,
        )

        execution.run_sync()
        controller.run_waiting_scripts(execution.stage_run)

        body, status, headers = execution.get_response()
        return flask.Response(status=status, headers=headers, response=body)

    @bp.route("/_jobs/<path:path>", methods=["POST"])
    def job_runner(path):
        if flask.request.headers.get("Shared-Token") != SIDECAR_SHARED_TOKEN:
            flask.abort(401)

        job = controller.get_job(path)
        if not job:
            flask.abort(404)

        if not job.file:
            flask.abort(500)

        request_data = RequestData(
            method=flask.request.method,
            body=flask.request.get_data(as_text=True),
            headers=flask.request.headers,
            query_params=flask.request.args,
        )

        def run_job(job):
            execution = JobExecution(
                is_initial=controller.is_initial(job.id),
                request=request_data,
                runtime_json=job,
            )

            execution.run_sync()
            controller.run_waiting_scripts(execution.stage_run)

        controller.executor.submit(run_job, job)

        return {"status": "running"}

    @bp.route("/", methods=["GET"])
    def index():
        res = send_from_dist("player.html", "player.html")
        return res

    @bp.route("/<path:filename>", methods=["GET"])
    def spa(filename: str):
        res = send_from_dist(filename, "player.html")
        return res

    return bp



from __future__ import annotations  # Required for TYPE_CHECKING
from typing import TYPE_CHECKING, Any, ClassVar, Dict, List, Optional
from dataclasses import dataclass
import threading, traceback, uuid, copy

from ..repositories.project.project import ProjectRepository
from ..repositories import StageRunRepository
from ..repositories.stage_run import StageRun
from ..utils.environment import IS_PREVIEW
from ..monitoring import LogMessage, log
from ..modules import import_as_new

if TYPE_CHECKING:
    from ..repositories.project.project import WorkflowStage


class NoExecutionFound(Exception):
    def __str__(self) -> str:
        return "No execution found. If you are running this locally, make sure you are running it with `abstra serve`. For more information, see https://docs.abstra.io"


@dataclass
class RequestData:
    headers: Dict[str, str]
    body: str
    method: str
    query_params: Dict[str, str]


class StageRunLockFailed(Exception):
    def __init__(self, status: Optional[str]) -> None:
        super().__init__(status)


class InvalidStageRunId(Exception):
    pass


class UnsetStageRun(Exception):
    pass


@dataclass
class WorkflowThread:
    _label: str
    _data: dict
    _assignee: Optional[str]
    _allowed_transitions: List[str]

    def set_transition(self, name: str):
        if name not in self._allowed_transitions:
            raise Exception(
                f"Transition '{name}' not found in {self._allowed_transitions}"
            )

        self._label = name

    def get_data(self) -> dict:
        return self._data

    def get_assignee(self) -> Optional[str]:
        return self._assignee

    def set_data(self, key: str, value: Any):
        self._data[key] = value

    def set_assignee(self, assignee: str):
        self._assignee = assignee

    def _clone(self) -> "WorkflowThread":
        return copy.deepcopy(self)


class Execution:
    type = "execution"
    abstra_run_key = "abstra-run-id"
    executions: ClassVar[Dict[int, "Execution"]] = {}

    id: str
    is_initial: bool
    request: RequestData
    thread: threading.Thread
    stage: WorkflowStage

    stage_run_freezed: Optional[StageRun] = None
    stage_run_draft: Optional[StageRun] = None
    workflow_threads: List[WorkflowThread]

    @classmethod
    def get_execution(cls) -> Optional["Execution"]:
        thread_id = threading.get_ident()
        return cls.executions.get(thread_id)

    def __init__(
        self,
        runtime_json: WorkflowStage,
        is_initial: bool,
        request: RequestData,
        execution_id=None,
    ):
        if execution_id:
            self.id = execution_id
        else:
            self.id = uuid.uuid4().__str__()

        self.is_initial = is_initial

        self.thread = threading.Thread(target=self._run, args=())
        self.request = request

        self.stderr: List[str] = []
        self.stdout: List[str] = []
        self.context: Dict = {}
        self.stage = runtime_json
        self.next_stage_runs = []

    @property
    def is_preview(self) -> bool:
        return IS_PREVIEW

    def run_async(self):
        self.thread.start()

    def run_sync(self):
        self.thread.start()
        self.thread.join()

    def stdio(self, type: str, text: str):
        if type == "stderr":
            self.stderr.append(text)
        elif type == "stdout":
            self.stdout.append(text)

    def log(self, event: str, payload: Any):
        if event == "heartbeat":
            return
        log(
            LogMessage(
                event=event,
                payload=payload,
                executionId=self.id,
                runtime_type=self.stage.runner_type,
                runtime_name=self.stage.title,
            )
        )

    def handle_started(self):
        pass

    def handle_success(self) -> str:
        return "finished"

    def handle_failure(self, e: Exception) -> str:
        traceback.print_exc()
        return "failed"

    def handle_lock_failed(self):
        stage_run_id = self.stage_run.id if self.stage_run else None
        stage_run_status = self.stage_run.status if self.stage_run else None

        self.log(
            "lock-failed",
            {
                "execution_id": self.id,
                "stage": self.stage.id,
                "stage_run_id": stage_run_id,
                "stage_run_status": stage_run_status,
            },
        )
        raise StageRunLockFailed(stage_run_status)

    def set_stage(
        self, data: Optional[Dict] = None, assignee: Optional[str] = None
    ) -> None:
        raise NotImplementedError("Can only be called from a hook")

    def setup_context(self, request: RequestData):
        raise NotImplementedError()

    @property
    def stage_run(self) -> Optional[StageRun]:
        return self.stage_run_draft

    @stage_run.setter
    def stage_run(self, stage_run: StageRun) -> None:
        self.stage_run_freezed = stage_run
        self.stage_run_draft = stage_run.clone()
        self.init_workflow_threads()

    def init_workflow_threads(self) -> None:
        if not self.stage_run:
            raise UnsetStageRun()

        allowed_transitions = [t.label for t in self.stage.workflow_transitions]
        self.workflow_threads = []
        if len(allowed_transitions) > 0:
            self.workflow_threads = [
                WorkflowThread(
                    _label=allowed_transitions[0],
                    _data=self.stage_run.data,
                    _assignee=self.stage_run.assignee,
                    _allowed_transitions=allowed_transitions,
                )
            ]

    def init_stage_run(self, stage_run_id: Optional[str] = None) -> None:
        if stage_run_id:
            stage_run = StageRunRepository.get(stage_run_id)
            if stage_run.stage == self.stage.id:
                self.stage_run = stage_run
                return
            else:
                raise InvalidStageRunId()

        if self.is_initial:
            self.stage_run = StageRunRepository.create_initial(stage=self.stage.id)
            return

        raise UnsetStageRun(
            f"This Stage is in the middle of a workflow, but no Step was specified"
        )

    def _run(self) -> None:
        self.thread_id = threading.get_ident()
        Execution.executions[self.thread_id] = self

        status = "running"
        self.setup_context(self.request)
        if self.stage_run and not self.set_stage_run_status(status):
            return self.handle_lock_failed()

        self.handle_started()
        try:
            try:
                import_as_new(self.stage.file)
            except SystemExit as e:
                if e.code is not None and e.code != 0:
                    raise e

            self.create_next_stage_runs()
            status = self.handle_success()
        except Exception as e:
            status = self.handle_failure(e)
        finally:
            self.delete()
            self.set_stage_run_status(status)

    def set_stage_run_status(self, status: str) -> bool:
        if not self.stage_run:
            return False

        return StageRunRepository.change_state(self.stage_run.id, status)

    def set_transition(self, transition_label: str) -> None:
        for transition in self.workflow_threads:
            transition.set_transition(transition_label)

    def set_next_assignee(self, assignee: str) -> None:
        for transition in self.workflow_threads:
            transition._assignee = assignee

    def set_next_data(self, key: str, value: Any) -> None:
        for transition in self.workflow_threads:
            transition._data[key] = value

    def split_workflow_thread(self, count: int) -> List[WorkflowThread]:
        if count <= 0:
            raise Exception("Count must be greater than 0")

        if len(self.workflow_threads) != 1:
            raise Exception("Thread was already splitted. You can't call this twice")

        transition = self.workflow_threads[0]

        for _ in range(count - 1):
            self.workflow_threads.append(transition._clone())

        return self.workflow_threads

    def create_next_stage_runs(self) -> None:
        if not self.stage_run:
            raise UnsetStageRun()

        project = ProjectRepository.load()  # TODO: inject this

        next_stage_runs_dtos = [
            dict(
                stage=project.follow_transition(self.stage.id, t._label).id,
                data=t._data,
                assignee=t._assignee,
            )
            for t in self.workflow_threads
        ]

        StageRunRepository.create_next(self.stage_run, next_stage_runs_dtos)

    def delete(self) -> None:
        del Execution.executions[self.thread_id]


def get_execution_throwable() -> Execution:
    execution = Execution.get_execution()
    if not execution:
        raise NoExecutionFound()
    return execution
